import sys
from pathlib import Path
from typing import Tuple

# Add parent directory to path for imports
current_dir = Path(__file__).parent
parent_dir = current_dir.parent.parent
sys.path.insert(0, str(parent_dir))
path = Path(__file__).parent
# ====================================================

import pygame
import spritePro as s


SPRITES_DIR = Path(__file__).resolve().parent / "Sprites"


def sprite_asset(name: str) -> str:
    return str((SPRITES_DIR / name).resolve())


def create_bar():
    return s.readySprites.BarWithBackground(
        sprite_asset("bar_bg.png"),
        sprite_asset("bar_fill.png"),
        (150, 30),
    )


class Hero(s.Sprite):
    def set_bar(self, bar: s.readySprites.BarWithBackground, pos: Tuple[int, int] = (0, -120)):
        self.bar = bar
        self.bar.set_parent(self)
        self.bar.local_offset = pos
        self.bar.set_fill_size((self.bar.size[0] // 1.02, self.bar.size[1] // 1.4))

    def update(self, screen: pygame.Surface = None):
        self.velocity = pygame.Vector2(self.velocity.x, 0)
        super().update(screen)


class HeroVsEnemyScene(s.Scene):
    def __init__(self):
        super().__init__()
        self.bg = s.Sprite(sprite_asset("background_game.png"), s.WH, s.WH_C, scene=self)
        self.bg.set_color((150, 150, 150))

        self.player = Hero(sprite_asset("hero.png"), speed=5, scene=self)
        self.player.set_native_size()
        self.player.set_position((s.WH_C.x, 730))
        self.player.set_scale(0.5)
        self.player.set_bar(create_bar())

        self.enemy = Hero(sprite_asset("enemy.png"), speed=1, scene=self)
        self.enemy.set_native_size()
        self.enemy.set_position((s.WH_C.x + 500, 730))
        self.enemy.set_scale(0.5)
        self.enemy.set_bar(create_bar())

    def update(self, dt: float) -> None:
        self.player.handle_keyboard_input(None, None, pygame.K_a, pygame.K_d)
        s.set_camera_follow(self.player, (0, -250))
        self.enemy.move_towards(self.player.get_position())


def run_demo(platform: str = "kivy") -> None:
    s.run(
        scene=HeroVsEnemyScene,
        size=(1260, 960),
        title="Game",
        fps=60,
        fill_color=(0, 0, 0),
        platform=platform,
    )


if __name__ == "__main__":
    # run_demo("pygame")
    run_demo()



import sys
from pathlib import Path
import random

import pygame

current_dir = Path(__file__).parent
parent_dir = current_dir.parent.parent
sys.path.insert(0, str(parent_dir))

import spritePro as s  # noqa: E402


class InputEventsDemoScene(s.Scene):
    def __init__(self):
        super().__init__()
        s.events.connect(s.globalEvents.KEY_DOWN, self.on_key_down)
        s.events.connect(s.globalEvents.QUIT, self.on_quit)

        self.player = s.Sprite("", (60, 60), (400, 300), speed=5, scene=self)
        self.player.set_color((120, 200, 255))
        self.player.angle = 0

        self.title = s.TextSprite(
            "Input + EventBus Demo", 28, (255, 255, 255), (400, 40), scene=self
        )
        self.hints = s.TextSprite(
            "R: color  |  Q: tween rotate 90  |  Space: scale",
            20,
            (200, 200, 200),
            (400, 560),
            scene=self,
        )

    def on_key_down(self, key, event):
        if key == pygame.K_ESCAPE:
            s.debug_log_info("ESC pressed (event)")

    def on_quit(self, event):
        s.debug_log_info("Quit requested")

    def update(self, dt: float) -> None:
        x_axis = s.input.get_axis(pygame.K_LEFT, pygame.K_RIGHT)
        y_axis = s.input.get_axis(pygame.K_UP, pygame.K_DOWN)
        self.player.velocity.x = x_axis * self.player.speed
        self.player.velocity.y = y_axis * self.player.speed

        if s.input.was_pressed(pygame.K_SPACE):
            self.player.set_scale(1.2)
        if s.input.was_released(pygame.K_SPACE):
            self.player.set_scale(1.0)
        if s.input.was_pressed(pygame.K_r):
            self.player.set_color(
                (
                    random.randint(80, 255),
                    random.randint(80, 255),
                    random.randint(80, 255),
                )
            )
        if s.input.was_pressed(pygame.K_q):
            start_angle = self.player.angle
            target_angle = start_angle + 90

            def apply_angle(value):
                self.player.rotate_to(value)

            tween = s.Tween(
                start_angle,
                target_angle,
                0.35,
                easing=s.EasingType.EASE_OUT,
                on_update=apply_angle,
            )
            tween.start()


def main():
    s.run(
        scene=InputEventsDemoScene,
        size=(800, 600),
        title="Input + EventBus Demo",
        fill_color=(20, 20, 30),
    )


if __name__ == "__main__":
    main()



"""Слайдер на базе Sprite."""

from __future__ import annotations

from typing import Callable, Optional, Tuple, TYPE_CHECKING

import pygame

from .sprite import Sprite
from .constants import Anchor

if TYPE_CHECKING:
    from .sprite import SpriteSceneInput


class Slider(Sprite):
    """Горизонтальный слайдер для UI на базе Sprite.

    Поддерживает два режима:
    - auto_register=True: участвует в игровом цикле, рисуется автоматически.
    - auto_register=False: не регистрируется; вызывайте handle_event() и draw() вручную.
    """

    def __init__(
        self,
        size: Tuple[int, int] = (200, 16),
        pos: Tuple[int, int] = (100, 100),
        min_value: float = 0.0,
        max_value: float = 1.0,
        value: float = 0.0,
        on_change: Optional[Callable[[float], None]] = None,
        on_release: Optional[Callable[[float], None]] = None,
        step: Optional[float] = None,
        track_color: Tuple[int, int, int] = (60, 60, 70),
        fill_color: Tuple[int, int, int] = (0, 150, 255),
        thumb_color: Tuple[int, int, int] = (220, 220, 220),
        sorting_order: int = 1000,
        auto_register: bool = True,
        scene: "SpriteSceneInput" = None,
    ):
        super().__init__(
            sprite="",
            size=size,
            pos=pos,
            sorting_order=sorting_order,
            anchor=Anchor.CENTER,
            scene=scene,
            auto_register=auto_register,
        )
        self.set_rect_shape(size, track_color, border_radius=4)
        self.set_screen_space(True)
        self.min_value = float(min_value)
        self.max_value = float(max_value) if max_value > min_value else float(min_value + 1.0)
        self.value = float(value)
        self.on_change = on_change
        self.on_release = on_release
        self.step = float(step) if step and step > 0 else None
        self.track_color = track_color
        self.fill_color = fill_color
        self.thumb_color = thumb_color
        self.dragging = False
        self._game_registered = auto_register
        self.set_value(self.value, emit=False)

    def set_rect(self, rect: pygame.Rect | Tuple[int, int, int, int]) -> "Slider":
        r = pygame.Rect(rect)
        self.set_position((r.centerx, r.centery), anchor=Anchor.CENTER)
        self.set_size((r.width, r.height))
        return self

    def set_range(self, min_value: float, max_value: float) -> "Slider":
        self.min_value = float(min_value)
        self.max_value = float(max_value) if max_value > min_value else float(min_value + 1.0)
        self.set_value(self.value, emit=False)
        return self

    def get_ratio(self) -> float:
        if self.max_value <= self.min_value:
            return 0.0
        ratio = (self.value - self.min_value) / (self.max_value - self.min_value)
        return max(0.0, min(1.0, ratio))

    def set_value(self, value: float, *, emit: bool = True) -> "Slider":
        clamped = max(self.min_value, min(self.max_value, float(value)))
        if self.step is not None:
            steps = round((clamped - self.min_value) / self.step)
            clamped = self.min_value + steps * self.step
            clamped = max(self.min_value, min(self.max_value, clamped))
        changed = abs(clamped - self.value) > 1e-9
        self.value = clamped
        if changed and emit and self.on_change is not None:
            self.on_change(self.value)
        return self

    def _value_from_x(self, x: float) -> float:
        r = self.rect
        if r.width <= 0:
            return self.min_value
        ratio = (x - r.left) / r.width
        ratio = max(0.0, min(1.0, ratio))
        return self.min_value + ratio * (self.max_value - self.min_value)

    def set_from_mouse_x(self, x: float, *, emit: bool = True) -> None:
        self.set_value(self._value_from_x(x), emit=emit)

    def handle_event(self, event: pygame.event.Event) -> bool:
        if (
            event.type == pygame.MOUSEBUTTONDOWN
            and event.button == 1
            and self.rect.collidepoint(event.pos)
        ):
            self.dragging = True
            self.set_from_mouse_x(event.pos[0], emit=True)
            return True
        if event.type == pygame.MOUSEBUTTONUP and event.button == 1 and self.dragging:
            self.dragging = False
            if self.on_release is not None:
                self.on_release(self.value)
            return True
        if event.type == pygame.MOUSEMOTION and self.dragging:
            self.set_from_mouse_x(event.pos[0], emit=True)
            return True
        return False

    def _render_image(self) -> pygame.Surface:
        w, h = self.size
        surf = pygame.Surface((w, h), pygame.SRCALPHA)
        rect = surf.get_rect()
        pygame.draw.rect(surf, self.track_color, rect, border_radius=4)
        ratio = self.get_ratio()
        fill_w = int(w * ratio)
        if fill_w > 0:
            pygame.draw.rect(
                surf,
                self.fill_color,
                pygame.Rect(0, 0, fill_w, h),
                border_radius=4,
            )
        thumb_x = int(w * ratio)
        thumb_rect = pygame.Rect(0, 0, 8, h + 4)
        thumb_rect.center = (thumb_x, h // 2)
        pygame.draw.rect(surf, self.thumb_color, thumb_rect, border_radius=3)
        return surf

    def update(self, screen: pygame.Surface = None):
        if not self.active:
            return
        try:
            import spritePro as s

            for ev in getattr(s, "pygame_events", []):
                if self.handle_event(ev):
                    break
        except Exception:
            pass
        surf = self._render_image()
        self.original_image = surf
        self._transformed_image = surf
        self._transform_dirty = False
        self._color_dirty = True
        super().update(screen)

    def draw(self, screen: pygame.Surface) -> None:
        """Ручная отрисовка (для auto_register=False)."""
        surf = self._render_image()
        r = surf.get_rect(center=self.rect.center)
        screen.blit(surf, r.topleft)



"""Текстовое поле ввода на базе Button."""

from __future__ import annotations

from typing import Callable, Optional, Tuple, TYPE_CHECKING

import pygame

from .button import Button
from .input_validation import (
    InputType,
    can_add_char,
    filter_chars_for_paste,
)

if TYPE_CHECKING:
    from .sprite import SpriteSceneInput


class TextInput(Button):
    """Поле ввода текста на базе Button без анимаций.

    Поддерживает типы: text (любой печатный текст), int (целые), float (дробные).
    Для int/float некорректные символы не вводятся и отфильтровываются при вставке.
    Enter — подтверждение, Escape — отмена. Ctrl+V — вставка, Ctrl+C — копирование содержимого поля.
    """

    def __init__(
        self,
        size: Tuple[int, int] = (200, 36),
        pos: Tuple[int, int] = (100, 100),
        placeholder: str = "",
        value: str = "",
        max_length: int = 128,
        input_type: InputType = "text",
        min_val: Optional[float] = None,
        max_val: Optional[float] = None,
        on_change: Optional[Callable[[str], None]] = None,
        on_submit: Optional[Callable[[str], None]] = None,
        text_color: Tuple[int, int, int] = (200, 200, 200),
        bg_color: Tuple[int, int, int] = (45, 45, 52),
        active_bg_color: Tuple[int, int, int] = (55, 55, 62),
        font_size: int = 18,
        sorting_order: int = 1000,
        scene: "SpriteSceneInput" = None,
    ):
        super().__init__(
            sprite="",
            size=size,
            pos=pos,
            text="",
            text_size=font_size,
            text_color=text_color,
            base_color=bg_color,
            hover_color=bg_color,
            press_color=bg_color,
            animated=False,
            use_scale_fx=False,
            use_color_fx=True,
            on_click=self._on_click_activate,
            sorting_order=sorting_order,
            scene=scene,
        )
        self.placeholder = placeholder
        self.value = value
        self.max_length = max(1, int(max_length))
        self.input_type: InputType = input_type
        self.min_val = min_val
        self.max_val = max_val
        self.on_change = on_change
        self.on_submit = on_submit
        self.is_active = False
        self._cursor_timer = 0.0
        self._show_cursor = True
        self._base_bg = bg_color
        self._active_bg = active_bg_color
        self._backspace_held_since: Optional[float] = None
        self._backspace_repeat_acc = 0.0
        self._backspace_cleared_all = False
        self._apply_text()

    def _on_click_activate(self) -> None:
        self.is_active = True
        self._cursor_timer = 0.0
        self._show_cursor = True
        self._apply_text()

    def _apply_text(self) -> None:
        shown = self.value if self.value else self.placeholder
        if self.is_active and self._show_cursor:
            shown = f"{self.value}|"
        self.text_sprite.set_text(shown)
        color = self._active_bg if self.is_active else self._base_bg
        self.set_all_colors(color, color, color)
        self.current_color = color

    def activate(self) -> None:
        self.is_active = True
        self._cursor_timer = 0.0
        self._show_cursor = True
        self._apply_text()

    def deactivate(self) -> None:
        self.is_active = False
        self._apply_text()

    def _paste_from_clipboard(self) -> None:
        try:
            if not pygame.scrap.get_init():
                pygame.scrap.init()
        except Exception:
            return
        try:
            data = pygame.scrap.get(pygame.SCRAP_TEXT)
            if not data:
                return
            text = data.decode("utf-8", errors="replace")
            paste = filter_chars_for_paste(self.input_type, text)
            if not paste:
                return
            remain = self.max_length - len(self.value)
            if remain <= 0:
                return
            self.value = (self.value + paste)[: self.max_length]
            self._apply_text()
            if self.on_change:
                self.on_change(self.value)
        except Exception:
            pass

    def _copy_to_clipboard(self) -> None:
        if not self.value:
            return
        try:
            if not pygame.scrap.get_init():
                pygame.scrap.init()
            pygame.scrap.put(pygame.SCRAP_TEXT, self.value.encode("utf-8"))
        except Exception:
            pass

    def set_value(self, value: str) -> None:
        self.value = value[: self.max_length]
        self._apply_text()
        if self.on_change:
            self.on_change(self.value)

    def handle_event(self, event: pygame.event.Event) -> bool:
        if event.type == pygame.MOUSEBUTTONDOWN and event.button == 1:
            if self.rect.collidepoint(event.pos):
                self.activate()
                return True
            if self.is_active:
                self.deactivate()
            return False
        if not self.is_active:
            return False
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_RETURN:
                self.deactivate()
                if self.on_submit:
                    self.on_submit(self.value)
                return True
            if event.key == pygame.K_ESCAPE:
                self.deactivate()
                return True
            if event.key == pygame.K_BACKSPACE:
                self.value = self.value[:-1]
                self._apply_text()
                if self.on_change:
                    self.on_change(self.value)
                self._backspace_held_since = 0.0
                self._backspace_repeat_acc = 0.0
                self._backspace_cleared_all = False
                return True
            mod = event.mod & (pygame.KMOD_CTRL | pygame.KMOD_META)
            if mod:
                if event.key == pygame.K_v:
                    self._paste_from_clipboard()
                    return True
                if event.key == pygame.K_c:
                    self._copy_to_clipboard()
                    return True
            return False
        if event.type == pygame.TEXTINPUT:
            text = event.text or ""
            buf = self.value
            for c in text:
                if len(buf) < self.max_length and can_add_char(self.input_type, buf, c):
                    buf += c
            if buf != self.value:
                self.value = buf
                self._apply_text()
                if self.on_change:
                    self.on_change(self.value)
            return True
        return False

    def update(self, screen: pygame.Surface = None):
        import spritePro as s

        for ev in getattr(s, "pygame_events", []):
            self.handle_event(ev)
        if self.is_active:
            self._cursor_timer += s.dt
            if self._cursor_timer >= 0.5:
                self._cursor_timer = 0.0
                self._show_cursor = not self._show_cursor
                self._apply_text()
            keys = pygame.key.get_pressed()
            if keys[pygame.K_BACKSPACE]:
                if self._backspace_held_since is None:
                    self._backspace_held_since = 0.0
                self._backspace_held_since += s.dt
                if self._backspace_held_since >= 3.0 and not self._backspace_cleared_all:
                    self.value = ""
                    self._apply_text()
                    if self.on_change:
                        self.on_change(self.value)
                    self._backspace_cleared_all = True
                elif self._backspace_held_since >= 1.0:
                    self._backspace_repeat_acc += s.dt
                    while self._backspace_repeat_acc >= 0.05 and self.value:
                        self.value = self.value[:-1]
                        self._backspace_repeat_acc -= 0.05
                        self._apply_text()
                        if self.on_change:
                            self.on_change(self.value)
            else:
                self._backspace_held_since = None
                self._backspace_repeat_acc = 0.0
                self._backspace_cleared_all = False
        super().update(screen)



"""Размещение кнопок и текста через лейауты spritePro."""

from __future__ import annotations

from typing import List

import pygame

from spritePro.layout import (
    LayoutAlignMain,
    LayoutAlignCross,
    layout_horizontal,
    layout_vertical,
)


class _Slot(pygame.sprite.Sprite):
    """Вспомогательный спрайт для расчёта позиций через Layout (только rect)."""

    def __init__(self, width: int, height: int) -> None:
        super().__init__()
        self.rect = pygame.Rect(0, 0, width, height)
        self.image = pygame.Surface((max(1, width), max(1, height)))
        self.image.set_alpha(0)

    def set_position(self, position, anchor=None) -> None:
        """Нужен для Layout: устанавливает rect по центру в position."""
        x, y = int(position[0]), int(position[1])
        self.rect.center = (x, y)


def row_rects(
    container: pygame.Rect,
    widths: List[int],
    gap: int = 6,
    align: str = "left",
    padding: int = 0,
) -> List[pygame.Rect]:
    """
    Горизонтальный ряд прямоугольников через spritePro.layout.
    align: "left", "center", "right".
    """
    if not widths:
        return []
    slots = [_Slot(w, container.height) for w in widths]
    align_main = {
        "left": LayoutAlignMain.START,
        "center": LayoutAlignMain.CENTER,
        "right": LayoutAlignMain.END,
    }.get(align, LayoutAlignMain.START)
    layout_horizontal(
        (container.x, container.y, container.width, container.height),
        slots,
        gap=gap,
        padding=padding,
        align_main=align_main,
        align_cross=LayoutAlignCross.CENTER,
        auto_apply=True,
    )
    return [s.rect.copy() for s in slots]


def column_rects(
    container: pygame.Rect,
    heights: List[int],
    gap: int = 4,
    padding: int = 0,
) -> List[pygame.Rect]:
    """Вертикальный столбец через spritePro.layout."""
    if not heights:
        return []
    slots = [_Slot(container.width, h) for h in heights]
    layout_vertical(
        (container.x, container.y, container.width, container.height),
        slots,
        gap=gap,
        padding=padding,
        align_main=LayoutAlignMain.START,
        align_cross=LayoutAlignCross.START,
        auto_apply=True,
    )
    return [s.rect.copy() for s in slots]


def pad(
    rect: pygame.Rect,
    left: int = 0,
    top: int = 0,
    right: int = 0,
    bottom: int = 0,
) -> pygame.Rect:
    """Внутренний прямоугольник с отступами."""
    return pygame.Rect(
        rect.x + left,
        rect.y + top,
        max(0, rect.width - left - right),
        max(0, rect.height - top - bottom),
    )



"""
Сборка проекта SpritePro для запуска в браузере (Pygbag / WebAssembly).

Использование через CLI: python -m spritePro.cli --webgl <путь_к_проекту>
Либо из кода: spritePro.web_build.build_web(...), build_web_archive(...).

При установке через pip (pip install spritepro) PACKAGE_ROOT указывает на
site-packages/spritePro/, откуда копируется пакет в папку сборки.
"""

from __future__ import annotations

import importlib.util
import json
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Any

# Корень пакета: в репо — spritePro/, при pip install — site-packages/spritePro/
PACKAGE_ROOT = Path(__file__).resolve().parent

DEFAULT_CONFIG = {
    "size": (800, 600),
    "title": "SpritePro Game",
    "fill_color": (20, 20, 30),
    "scene_import": "from scene import MainScene",
    "scene_var": "MainScene",
    "py_files": ["scene.py", "config.py"],
    "extra_setup": "",
}

_CONFIG_SIZE_KEYS = ("WINDOW_SIZE", "SIZE", "SCREEN_SIZE", "GAME_SIZE")
_CONFIG_TITLE_KEYS = ("TITLE", "GAME_TITLE", "WINDOW_TITLE")
_CONFIG_FILL_COLOR_KEYS = ("FILL_COLOR", "BACKGROUND_COLOR", "BG_COLOR")


def _ignore_for_web(
    directory: Path,
    names: list,
    *,
    skip_dir_name: str | None = None,
    skip_dir_parent: Path | None = None,
) -> list:
    result = [
        n
        for n in names
        if n == "__pycache__" or n.endswith((".pyc", ".pyo", ".mp3", ".wav")) or n.startswith(".")
    ]
    if (
        skip_dir_name
        and skip_dir_parent is not None
        and Path(directory).resolve() == skip_dir_parent.resolve()
    ):
        result.append(skip_dir_name)
    return result


def _infer_from_config_module(
    project_dir: Path, cfg: dict[str, Any], module_name: str = "config"
) -> None:
    """Подтягивает size, title, fill_color из модуля конфига игры, если они там заданы."""
    module_file = project_dir / f"{module_name}.py"
    if not module_file.is_file():
        return
    try:
        spec = importlib.util.spec_from_file_location(module_name, module_file)
        if spec is None or spec.loader is None:
            return
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        for key in _CONFIG_SIZE_KEYS:
            val = getattr(mod, key, None)
            if val is not None and isinstance(val, (tuple, list)) and len(val) >= 2:
                cfg["size"] = (int(val[0]), int(val[1]))
                break
        for key in _CONFIG_TITLE_KEYS:
            val = getattr(mod, key, None)
            if val is not None and isinstance(val, str):
                cfg["title"] = val
                break
        for key in _CONFIG_FILL_COLOR_KEYS:
            val = getattr(mod, key, None)
            if val is not None and isinstance(val, (tuple, list)) and len(val) >= 3:
                cfg["fill_color"] = (int(val[0]), int(val[1]), int(val[2]))
                break
    except Exception:
        pass


def load_config(project_dir: Path) -> dict[str, Any]:
    """Загружает конфиг из webgl.json; при отсутствии size/title/fill_color подтягивает из config.py игры."""
    cfg = DEFAULT_CONFIG.copy()
    data: dict[str, Any] = {}
    config_path = project_dir / "webgl.json"
    if config_path.is_file():
        try:
            data = json.loads(config_path.read_text(encoding="utf-8"))
            if "size" in data:
                cfg["size"] = tuple(data["size"])
            if "title" in data:
                cfg["title"] = data["title"]
            if "fill_color" in data:
                cfg["fill_color"] = tuple(data["fill_color"])
            if "scene_import" in data:
                cfg["scene_import"] = data["scene_import"]
            if "scene_var" in data:
                cfg["scene_var"] = data["scene_var"]
            if "py_files" in data:
                cfg["py_files"] = list(data["py_files"])
            if "extra_setup" in data:
                cfg["extra_setup"] = data["extra_setup"]
        except (json.JSONDecodeError, OSError) as e:
            sys.stderr.write(f"Предупреждение: не удалось прочитать {config_path}: {e}\n")
    config_module = data.get("config_module", "config")
    _infer_from_config_module(project_dir, cfg, config_module)
    return cfg


def build_web(
    project_dir: Path,
    output_dir: Path | None = None,
    config: dict[str, Any] | None = None,
) -> Path:
    """
    Подготавливает папку для веб-сборки Pygbag.

    Копирует пакет spritePro и указанные файлы игры, создаёт main.py и game_entry.py.

    Args:
        project_dir: Папка проекта (содержит scene.py, config.py и опционально webgl.json).
        output_dir: Папка вывода. По умолчанию project_dir / "web_build".
        config: Конфиг сборки. Если None, загружается из project_dir / "webgl.json" или дефолт.

    Returns:
        Путь к папке сборки (output_dir).
    """
    if output_dir is None:
        output_dir = project_dir / "web_build"
    output_dir = output_dir.resolve()
    if config is None:
        config = load_config(project_dir)

    size = config["size"]
    title = config["title"]
    fill_color = config["fill_color"]
    scene_import = config["scene_import"]
    scene_var = config["scene_var"]
    py_files = config["py_files"]
    extra_setup = config.get("extra_setup", "")

    output_dir.mkdir(parents=True, exist_ok=True)

    # Чтобы не копировать саму папку сборки, если она внутри пакета (например demoGames/ping_pong/web_build)
    skip_web_build = None
    try:
        if output_dir.resolve().relative_to(PACKAGE_ROOT):
            skip_web_build = output_dir.name
    except ValueError:
        pass

    def ignore(directory: Path, names: list) -> list:
        return _ignore_for_web(
            directory,
            names,
            skip_dir_name=skip_web_build,
            skip_dir_parent=output_dir.parent if skip_web_build else None,
        )

    # Копируем spritePro
    spritepro_dst = output_dir / "spritePro"
    if spritepro_dst.exists():
        shutil.rmtree(spritepro_dst)
    shutil.copytree(PACKAGE_ROOT, spritepro_dst, ignore=ignore)

    # Копируем файлы игры
    for fname in py_files:
        src = project_dir / fname
        if src.is_file():
            shutil.copy2(src, output_dir / fname)

    # game_entry.py
    extra_line = f"\n    {extra_setup}" if extra_setup else ""
    game_entry_py = f'''"""Точка входа для веб-сборки (генерируется spritePro.web_build)."""
import spritePro as s
{scene_import}


def setup() -> None:
    s.get_screen({size!r}, {title!r}){extra_line}
    s.set_scene({scene_var}())
'''
    (output_dir / "game_entry.py").write_text(game_entry_py, encoding="utf-8")

    # main.py
    r, g, b = fill_color
    main_py = f'''"""Async-вход для Pygbag (генерируется spritePro.web_build)."""
import asyncio
import sys
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parent))

import spritePro as s
import game_entry


async def main() -> None:
    game_entry.setup()
    while True:
        s.update(fill_color=({r}, {g}, {b}))
        await asyncio.sleep(0)
        if s.quit_requested():
            break


if __name__ == "__main__":
    asyncio.run(main())
'''
    (output_dir / "main.py").write_text(main_py, encoding="utf-8")

    return output_dir


def build_web_archive(
    project_dir: Path,
    output_dir: Path | None = None,
    config: dict[str, Any] | None = None,
    zip_path: Path | None = None,
) -> Path:
    """
    Подготавливает веб-сборку и создаёт ZIP-архив для публикации (Яндекс Игры, itch.io и др.).

    Вызывает build_web(), затем pygbag --build --archive. В архиве в корне лежит index.html
    и все файлы билда — формат подходит для загрузки на площадки HTML5-игр.

    Args:
        project_dir: Папка проекта с webgl.json и файлами игры.
        output_dir: Папка веб-сборки. По умолчанию project_dir / "web_build".
        config: Конфиг. Если None — из webgl.json или дефолт.
        zip_path: Куда сохранить ZIP. По умолчанию <output_dir>/build/web.zip (результат pygbag).

    Returns:
        Путь к созданному ZIP-файлу.

    Raises:
        FileNotFoundError: Если pygbag не установлен (pip install spritepro[web]).
        RuntimeError: Если pygbag завершился с ошибкой.
    """
    out = build_web(project_dir, output_dir=output_dir, config=config)
    try:
        proc = subprocess.run(
            [sys.executable, "-m", "pygbag", "--build", "--archive", str(out)],
            capture_output=True,
            text=True,
            timeout=300,
        )
    except FileNotFoundError as e:
        raise FileNotFoundError(
            "Для сборки архива нужен pygbag. Установите: pip install spritepro[web]"
        ) from e
    if proc.returncode != 0:
        raise RuntimeError(
            f"pygbag завершился с кодом {proc.returncode}. stderr: {proc.stderr or proc.stdout}"
        )
    if zip_path and zip_path.is_file():
        return zip_path
    build_dir = out / "build"
    if zip_path:
        archive_file = zip_path
    elif (build_dir / "web.zip").is_file():
        archive_file = build_dir / "web.zip"
    else:
        zips = list(build_dir.glob("*.zip"))
        if not zips:
            raise FileNotFoundError(f"Архив не найден в {build_dir}. Проверьте вывод pygbag.")
        archive_file = zips[0]
    return archive_file



"""Demo: Visibility Culling — проверка is_visible_on_screen().

Спрайты движутся и отскакивают от границ области. Счётчики показывают,
сколько из них попадают в текущий вид камеры (Visible) и сколько вне (Offscreen).
Двигайте камеру WASD — числа меняются. Спрайты не исчезают, они всегда в области и периодически заходят в кадр.
"""

import pygame
import spritePro as s

WORLD_LEFT = -80
WORLD_RIGHT = 880
WORLD_TOP = -80
WORLD_BOTTOM = 680


def run_demo():
    s.init()
    s.get_screen((800, 600), "Visibility Culling Demo")
    s.enable_debug()
    s.debug_log_info("Visibility Culling: WASD — камера, спрайты отскакивают в области")

    sprites = []
    num_sprites = 60

    for i in range(num_sprites):
        x = 100 + (i % 10) * 70
        y = 100 + (i // 10) * 80
        sprite = s.Sprite("", pos=(x, y), size=(36, 36))
        colors = [(255, 100, 100), (100, 255, 100), (100, 100, 255), (255, 255, 100)]
        sprite.set_circle_shape(radius=18, color=colors[i % 4])
        angle_deg = i * 37
        v = pygame.math.Vector2(1, 0).rotate(angle_deg) * 90
        sprite.velocity = pygame.math.Vector2(v.x, v.y)
        sprites.append(sprite)

    visible_text = s.TextSprite("Visible: 0", color=(0, 255, 0), pos=(20, 30))
    visible_text.set_position((20, 30), anchor="topleft")
    visible_text.set_screen_space(True)
    offscreen_text = s.TextSprite("Offscreen: 0", color=(255, 150, 150), pos=(20, 52))
    offscreen_text.set_position((20, 52), anchor="topleft")
    offscreen_text.set_screen_space(True)
    hint = s.TextSprite(
        "WASD: move camera — visible count changes", color=(200, 200, 200), pos=(20, 78)
    )
    hint.set_position((20, 78), anchor="topleft")
    hint.set_screen_space(True)

    while True:
        s.update(fill_color=(20, 20, 35))

        if s.input.is_pressed(pygame.K_a):
            s.move_camera(-280 * s.dt, 0)
        if s.input.is_pressed(pygame.K_d):
            s.move_camera(280 * s.dt, 0)
        if s.input.is_pressed(pygame.K_w):
            s.move_camera(0, -280 * s.dt)
        if s.input.is_pressed(pygame.K_s):
            s.move_camera(0, 280 * s.dt)

        visible_count = 0
        for sprite in sprites:
            sprite.update()
            if sprite.rect.left < WORLD_LEFT:
                sprite.rect.left = WORLD_LEFT
                sprite.velocity.x = abs(sprite.velocity.x)
            if sprite.rect.right > WORLD_RIGHT:
                sprite.rect.right = WORLD_RIGHT
                sprite.velocity.x = -abs(sprite.velocity.x)
            if sprite.rect.top < WORLD_TOP:
                sprite.rect.top = WORLD_TOP
                sprite.velocity.y = abs(sprite.velocity.y)
            if sprite.rect.bottom > WORLD_BOTTOM:
                sprite.rect.bottom = WORLD_BOTTOM
                sprite.velocity.y = -abs(sprite.velocity.y)

            if sprite.is_visible_on_screen(margin=20):
                visible_count += 1

        offscreen_count = num_sprites - visible_count
        visible_text.set_text(f"Visible: {visible_count}")
        offscreen_text.set_text(f"Offscreen: {offscreen_count}")


if __name__ == "__main__":
    run_demo()



"""Object pooling utilities for SpritePro."""

from __future__ import annotations

from typing import Callable, Generic, Optional, TypeVar, Any
from collections import deque

T = TypeVar("T")


class ObjectPool(Generic[T]):
    """Пул объектов для переиспользования экземпляров и снижения нагрузки на GC.

    Attributes:
        factory (Callable[[], T]): Фабрика для создания новых объектов.
        reset_fn (Optional[Callable[[T], None]]): Функция сброса объекта при возврате в пул.
        max_size (int): Максимальный размер пула (0 = без ограничений).
        initial_size (int): Начальное количество объектов в пуле.

    Example:
        >>> pool = ObjectPool(lambda: MySprite(), reset_fn=lambda s: s.reset())
        >>> obj = pool.acquire()
        >>> pool.release(obj)
    """

    def __init__(
        self,
        factory: Callable[[], T],
        reset_fn: Optional[Callable[[T], None]] = None,
        max_size: int = 0,
        initial_size: int = 0,
    ):
        """Инициализирует пул объектов.

        Args:
            factory (Callable[[], T]): Фабрика для создания новых объектов.
            reset_fn (Optional[Callable[[T], None]], optional): Функция сброса при возврате. По умолчанию None.
            max_size (int, optional): Максимальный размер пула (0 = без ограничений). По умолчанию 0.
            initial_size (int, optional): Начальное количество объектов. По умолчанию 0.
        """
        self.factory = factory
        self.reset_fn = reset_fn
        self.max_size = max_size
        self._pool: deque[T] = deque()
        self._active_count: int = 0
        self._total_created: int = 0

        if initial_size > 0:
            self._prefill(initial_size)

    def _prefill(self, count: int) -> None:
        """Заполняет пул предсозданными объектами."""
        for _ in range(count):
            self._pool.append(self.factory())
            self._total_created += 1

    def acquire(self) -> T:
        """Получает объект из пула.

        Returns:
            T: Объект из пула или новый объект.

        Note:
            Если пул пуст, создаётся новый объект через фабрику.
        """
        self._active_count += 1

        if self._pool:
            obj = self._pool.popleft()
        else:
            obj = self.factory()
            self._total_created += 1

        return obj

    def release(self, obj: T) -> bool:
        """Возвращает объект в пул.

        Args:
            obj (T): Объект для возврата.

        Returns:
            bool: True если объект принят в пул, False если отклонён (достигнут максимальный размер).
        """
        self._active_count -= 1

        if self.reset_fn:
            self.reset_fn(obj)

        if self.max_size > 0 and len(self._pool) >= self.max_size:
            return False

        self._pool.append(obj)
        return True

    def clear(self) -> None:
        """Очищает пул, удаляя все неактивные объекты."""
        self._pool.clear()

    @property
    def size(self) -> int:
        """Количество объектов в пуле (неактивных)."""
        return len(self._pool)

    @property
    def active_count(self) -> int:
        """Количество активных (используемых) объектов."""
        return self._active_count

    @property
    def total_created(self) -> int:
        """Общее количество созданных объектов (включая те, что были отклонены при max_size)."""
        return self._total_created


class PooledSpritePool(ObjectPool):
    """Специализированный пул для спрайтов SpritePro.

    Предоставляет методы для создания и сброса спрайтов с поддержкой
    всех параметров конструктора Sprite.
    """

    def __init__(
        self,
        sprite_class: type,
        max_size: int = 100,
        initial_size: int = 0,
        **default_kwargs: Any,
    ):
        """Инициализирует пул спрайтов.

        Args:
            sprite_class (type): Класс спрайта (должен быть подклассом Sprite).
            max_size (int, optional): Максимальный размер пула. По умолчанию 100.
            initial_size (int, optional): Начальное количество объектов. По умолчанию 0.
            **default_kwargs: Параметры по умолчанию для создания спрайтов.
        """
        self.sprite_class = sprite_class
        self.default_kwargs = default_kwargs

        def create_sprite() -> T:
            return sprite_class(**default_kwargs)

        def reset_sprite(sprite: T) -> None:
            if hasattr(sprite, "kill"):
                sprite.kill()
            if hasattr(sprite, "reset_sprite"):
                sprite.reset_sprite(**default_kwargs)
            elif hasattr(sprite, "_pool_release"):
                sprite._pool_release = None

        super().__init__(create_sprite, reset_sprite, max_size, initial_size)


class PoolManager:
    """Глобальный менеджер пулов для централизованного управления.

    Позволяет регистрировать и получать пулы по имени, автоматически
    создавая их при первом обращении.
    """

    _instance: Optional["PoolManager"] = None
    _pools: dict[str, ObjectPool] = {}

    def __new__(cls) -> "PoolManager":
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._pools = {}
        return cls._instance

    @classmethod
    def get(cls, name: str) -> Optional[ObjectPool]:
        """Получает пул по имени."""
        return cls._instance._pools.get(name)

    @classmethod
    def register(cls, name: str, pool: ObjectPool) -> None:
        """Регистрирует пул под именем."""
        cls._instance._pools[name] = pool

    @classmethod
    def create_pool(
        cls,
        name: str,
        factory: Callable[[], Any],
        reset_fn: Optional[Callable[[Any], None]] = None,
        max_size: int = 0,
    ) -> ObjectPool:
        """Создаёт и регистрирует пул.

        Returns:
            ObjectPool: Созданный пул.
        """
        pool = ObjectPool(factory, reset_fn, max_size)
        cls.register(name, pool)
        return pool

    @classmethod
    def clear_all(cls) -> None:
        """Очищает все пулы."""
        for pool in cls._instance._pools.values():
            pool.clear()


def pool_acquire(pool_name: str) -> Any:
    """Получает объект из именованного пула.

    Args:
        pool_name (str): Имя пула.

    Returns:
        Any: Объект из пула.

    Raises:
        KeyError: Если пул с указанным именем не найден.
    """
    pool = PoolManager.get(pool_name)
    if pool is None:
        raise KeyError(f"Pool '{pool_name}' not found")
    return pool.acquire()


def pool_release(pool_name: str, obj: Any) -> bool:
    """Возвращает объект в именованный пул.

    Args:
        pool_name (str): Имя пула.
        obj (Any): Объект для возврата.

    Returns:
        bool: True если объект принят.

    Raises:
        KeyError: Если пул не найден.
    """
    pool = PoolManager.get(pool_name)
    if pool is None:
        raise KeyError(f"Pool '{pool_name}' not found")
    return pool.release(obj)



from __future__ import annotations

import logging
import os
import subprocess
import sys
from pathlib import Path


SCREEN_PRESETS: dict[str, tuple[int, int]] = {
    "phone-portrait": (360, 640),
    "phone-tall": (412, 915),
    "phone-landscape": (640, 360),
    "tablet-landscape": (1280, 720),
    "tablet-portrait": (800, 1280),
}


def preview_target_from_path(target: Path) -> Path:
    resolved = target.resolve()
    if resolved.is_file() and resolved.suffix == ".py":
        return resolved
    if resolved.is_dir():
        main_file = resolved / "main.py"
        if main_file.is_file():
            return main_file
    raise FileNotFoundError(f"Не удалось найти main.py или указанный .py файл: {target}")


def parse_preview_size(value: str | None) -> tuple[int, int] | None:
    if not value:
        return None
    normalized = value.lower().replace(" ", "")
    separator = "x" if "x" in normalized else ","
    if separator not in normalized:
        raise ValueError("Размер должен быть в формате WIDTHxHEIGHT, например 360x640")
    width_str, height_str = normalized.split(separator, 1)
    width = max(1, int(width_str))
    height = max(1, int(height_str))
    return width, height


def run_preview(
    target: Path,
    *,
    platform: str,
    preset: str | None,
    size: str | None,
    extra_args: list[str],
) -> int:
    main_file = preview_target_from_path(target)
    if preset and size:
        raise ValueError("Используйте либо --screen, либо --size, но не оба сразу")

    resolved_size = parse_preview_size(size)
    if preset:
        resolved_size = SCREEN_PRESETS[preset]
    if resolved_size is None:
        resolved_size = SCREEN_PRESETS["phone-portrait"]

    env = os.environ.copy()
    env["SPRITEPRO_WINDOW_SIZE"] = f"{resolved_size[0]}x{resolved_size[1]}"
    env["SPRITEPRO_PLATFORM"] = platform
    env["SPRITEPRO_TITLE_SUFFIX"] = f"[{platform} {resolved_size[0]}x{resolved_size[1]}]"

    logging.info("Preview target: %s", main_file)
    logging.info("Platform: %s", platform)
    logging.info("Window size: %sx%s", resolved_size[0], resolved_size[1])
    if extra_args:
        logging.info("Forward args: %s", " ".join(extra_args))

    completed = subprocess.run([sys.executable, str(main_file), *extra_args], env=env)
    return completed.returncode



from __future__ import annotations

from typing import Dict, Iterable, Set, Tuple

import pygame


class InputState:
    """Состояние ввода в стиле Unity для клавиатуры и мыши.

    Хранит события текущего кадра и даёт быстрые методы проверки:
    удержание, нажатие и отпускание клавиш/кнопок.
    """

    def __init__(self) -> None:
        """Инициализирует состояние ввода и буферы событий."""
        self._keys_down: Set[int] = set()
        self._keys_up: Set[int] = set()
        self._keys_pressed_state: Dict[int, bool] = {}
        self._keys_pressed: Tuple[bool, ...] = tuple()

        self._mouse_down: Set[int] = set()
        self._mouse_up: Set[int] = set()
        self._mouse_pressed: Tuple[bool, ...] = tuple()
        self._mouse_buttons_state: Dict[int, bool] = {}
        self.mouse_pos: Tuple[int, int] = (0, 0)
        self.mouse_rel: Tuple[int, int] = (0, 0)
        self.mouse_wheel: Tuple[int, int] = (0, 0)
        self._last_mouse_pos: Tuple[int, int] = (0, 0)

    def update(
        self,
        events: Iterable[pygame.event.Event],
        *,
        poll_hardware: bool = True,
    ) -> None:
        """Обновляет состояние ввода по событиям текущего кадра.

        Args:
            events (Iterable[pygame.event.Event]): События pygame за кадр.
        """
        if isinstance(events, list):
            events_list = events
        else:
            events_list = list(events)
        self._keys_down.clear()
        self._keys_up.clear()
        self._mouse_down.clear()
        self._mouse_up.clear()
        self.mouse_wheel = (0, 0)
        self.mouse_rel = (0, 0)
        self._last_mouse_pos = self.mouse_pos

        saw_mouse_event = False

        for event in events_list:
            if event.type == pygame.KEYDOWN:
                self._keys_down.add(event.key)
                self._keys_pressed_state[event.key] = True
            elif event.type == pygame.KEYUP:
                self._keys_up.add(event.key)
                self._keys_pressed_state[event.key] = False
            elif event.type == pygame.MOUSEBUTTONDOWN:
                self._mouse_down.add(event.button)
                self._mouse_buttons_state[event.button] = True
                if hasattr(event, "pos"):
                    self.mouse_pos = (int(event.pos[0]), int(event.pos[1]))
                    saw_mouse_event = True
            elif event.type == pygame.MOUSEBUTTONUP:
                self._mouse_up.add(event.button)
                self._mouse_buttons_state[event.button] = False
                if hasattr(event, "pos"):
                    self.mouse_pos = (int(event.pos[0]), int(event.pos[1]))
                    saw_mouse_event = True
            elif event.type == pygame.MOUSEMOTION:
                if hasattr(event, "pos"):
                    self.mouse_pos = (int(event.pos[0]), int(event.pos[1]))
                    saw_mouse_event = True
                if hasattr(event, "rel"):
                    self.mouse_rel = (int(event.rel[0]), int(event.rel[1]))
            elif event.type == pygame.MOUSEWHEEL:
                self.mouse_wheel = (event.x, event.y)

        if poll_hardware:
            self._keys_pressed = pygame.key.get_pressed()
            try:
                mouse_pressed = pygame.mouse.get_pressed(5)
            except TypeError:
                mouse_pressed = pygame.mouse.get_pressed()
            if mouse_pressed:
                self._mouse_pressed = tuple(bool(v) for v in mouse_pressed)
                for index, value in enumerate(self._mouse_pressed, start=1):
                    self._mouse_buttons_state[index] = bool(value)
        else:
            self._keys_pressed = tuple()
            self._mouse_pressed = tuple()

        if not saw_mouse_event and poll_hardware:
            try:
                current_mouse_pos = pygame.mouse.get_pos()
                self.mouse_pos = (int(current_mouse_pos[0]), int(current_mouse_pos[1]))
                rel = pygame.mouse.get_rel()
                self.mouse_rel = (int(rel[0]), int(rel[1]))
            except pygame.error:
                self.mouse_rel = (
                    self.mouse_pos[0] - self._last_mouse_pos[0],
                    self.mouse_pos[1] - self._last_mouse_pos[1],
                )
        elif self.mouse_rel == (0, 0):
            self.mouse_rel = (
                self.mouse_pos[0] - self._last_mouse_pos[0],
                self.mouse_pos[1] - self._last_mouse_pos[1],
            )

    def is_pressed(self, key: int) -> bool:
        """Проверяет, удерживается ли клавиша.

        Args:
            key (int): Код клавиши pygame.

        Returns:
            bool: True, если клавиша удерживается.
        """
        if key in self._keys_pressed_state:
            return self._keys_pressed_state[key]
        try:
            return bool(self._keys_pressed[key])
        except IndexError:
            return False

    def was_pressed(self, key: int) -> bool:
        """Проверяет нажатие клавиши в текущем кадре.

        Args:
            key (int): Код клавиши pygame.

        Returns:
            bool: True, если клавиша нажата в этом кадре.
        """
        return key in self._keys_down

    def was_released(self, key: int) -> bool:
        """Проверяет отпускание клавиши в текущем кадре.

        Args:
            key (int): Код клавиши pygame.

        Returns:
            bool: True, если клавиша отпущена в этом кадре.
        """
        return key in self._keys_up

    def get_axis(self, negative_key: int, positive_key: int) -> int:
        """Возвращает значение оси -1/0/1 по двум клавишам.

        Args:
            negative_key (int): Клавиша для отрицательного направления.
            positive_key (int): Клавиша для положительного направления.

        Returns:
            int: -1, 0 или 1.
        """
        neg = 1 if self.is_pressed(negative_key) else 0
        pos = 1 if self.is_pressed(positive_key) else 0
        return pos - neg

    def is_mouse_pressed(self, button: int) -> bool:
        """Проверяет удержание кнопки мыши.

        Args:
            button (int): Кнопка мыши (1=ЛКМ, 2=СКМ, 3=ПКМ).

        Returns:
            bool: True, если кнопка удерживается.
        """
        if button in self._mouse_buttons_state:
            return self._mouse_buttons_state[button]
        index = max(0, min(button - 1, len(self._mouse_pressed) - 1))
        return bool(self._mouse_pressed[index]) if self._mouse_pressed else False

    def was_mouse_pressed(self, button: int) -> bool:
        """Проверяет нажатие кнопки мыши в текущем кадре.

        Args:
            button (int): Кнопка мыши (1=ЛКМ, 2=СКМ, 3=ПКМ).

        Returns:
            bool: True, если кнопка нажата в этом кадре.
        """
        return button in self._mouse_down

    def was_mouse_released(self, button: int) -> bool:
        """Проверяет отпускание кнопки мыши в текущем кадре.

        Args:
            button (int): Кнопка мыши (1=ЛКМ, 2=СКМ, 3=ПКМ).

        Returns:
            bool: True, если кнопка отпущена в этом кадре.
        """
        return button in self._mouse_up



"""Обработка ввода: текстовые поля с типами text / int / float (редактор)."""

from typing import Optional

import pygame
import string

from ...input_validation import (
    InputType,
    can_add_char as _can_add_char,
    filter_chars_for_paste as _filter_chars_for_paste,
    parse_input_value as _parse_input_value,
)

ALLOWED_NAME_CHARS = set(string.ascii_letters + string.digits + string.whitespace + "._-()")

KEYPAD_MAP = {
    pygame.K_KP0: "0",
    pygame.K_KP1: "1",
    pygame.K_KP2: "2",
    pygame.K_KP3: "3",
    pygame.K_KP4: "4",
    pygame.K_KP5: "5",
    pygame.K_KP6: "6",
    pygame.K_KP7: "7",
    pygame.K_KP8: "8",
    pygame.K_KP9: "9",
    pygame.K_KP_PERIOD: ".",
    pygame.K_KP_MINUS: "-",
    pygame.K_MINUS: "-",
    pygame.K_PERIOD: ".",
    pygame.K_COMMA: ",",
}


def _allowed_for_editor_text(name: str) -> Optional[set]:
    return ALLOWED_NAME_CHARS if name == "prop_input_name" else None


def can_add_char(input_type: InputType, current: str, ch: str, name: str = "") -> bool:
    if name == "prop_input_text" and ch == "\n":
        return input_type == "text"
    return _can_add_char(input_type, current, ch, _allowed_for_editor_text(name) if name else None)


def filter_chars_for_paste(input_type: InputType, text: str, name: str = "") -> str:
    if input_type == "text" and name == "prop_input_text":
        normalized = text.replace("\r\n", "\n").replace("\r", "\n")
        return "".join(c for c in normalized if c == "\n" or _can_add_char(input_type, "", c, None))
    return _filter_chars_for_paste(
        input_type, text, _allowed_for_editor_text(name) if name else None
    )


def parse_input_value(
    input_type: InputType,
    raw: str,
    min_val: Optional[float] = None,
    max_val: Optional[float] = None,
):
    return _parse_input_value(input_type, raw, min_val, max_val)


def get_active_input_type(editor) -> InputType:
    return getattr(editor, "_active_text_input_type", "text")


def _paste_into_editor_buffer(editor, name: str) -> None:
    try:
        if not pygame.scrap.get_init():
            pygame.scrap.init()
    except Exception:
        return
    try:
        data = pygame.scrap.get(pygame.SCRAP_TEXT)
        if not data:
            return
        text = data.decode("utf-8", errors="replace")
        input_type = get_active_input_type(editor)
        filtered = filter_chars_for_paste(input_type, text, name)
        if filtered:
            editor._text_input_buffers[name] = editor._text_input_buffers.get(name, "") + filtered
    except Exception:
        pass


def _copy_editor_buffer_to_clipboard(editor, name: str) -> None:
    try:
        if not pygame.scrap.get_init():
            pygame.scrap.init()
    except Exception:
        return
    try:
        text = editor._text_input_buffers.get(name, "")
        if text:
            pygame.scrap.put(pygame.SCRAP_TEXT, text.encode("utf-8"))
    except Exception:
        pass


def handle_text_input_keydown(editor, event: pygame.event.Event) -> bool:
    """Обработка KEYDOWN при активном текстовом поле. Возвращает True если событие обработано."""
    name = editor._active_text_input
    if name is None:
        return False

    mod = event.mod & (pygame.KMOD_CTRL | pygame.KMOD_META)
    if mod:
        if event.key == pygame.K_v:
            _paste_into_editor_buffer(editor, name)
            return True
        if event.key == pygame.K_c:
            _copy_editor_buffer_to_clipboard(editor, name)
            return True

    if event.key == pygame.K_RETURN:
        if name == "prop_input_text" and not mod:
            editor._text_input_buffers[name] = editor._text_input_buffers.get(name, "") + "\n"
            return True
        editor._deactivate_text_input(apply=True)
        return True
    if event.key == pygame.K_ESCAPE:
        editor._deactivate_text_input(apply=False)
        return True
    if event.key == pygame.K_BACKSPACE:
        editor._text_input_buffers[name] = editor._text_input_buffers[name][:-1]
        return True

    input_type = get_active_input_type(editor)
    if input_type == "int":
        keypad_no_dot = {k: v for k, v in KEYPAD_MAP.items() if v not in ".,"}
        if event.key in keypad_no_dot and not (event.unicode or ""):
            ch = keypad_no_dot[event.key]
            buf = editor._text_input_buffers.get(name, "")
            if can_add_char(input_type, buf, ch, name):
                editor._text_input_buffers[name] = buf + ch
            return True
    elif event.key in KEYPAD_MAP and not (event.unicode or ""):
        ch = KEYPAD_MAP[event.key]
        buf = editor._text_input_buffers.get(name, "")
        if can_add_char(input_type, buf, ch, name):
            editor._text_input_buffers[name] = buf + ch
        return True
    return True


def handle_text_input_text(editor, event: pygame.event.Event) -> bool:
    """Обработка TEXTINPUT при активном текстовом поле."""
    name = editor._active_text_input
    if name is None:
        return False
    input_type = get_active_input_type(editor)
    buf = editor._text_input_buffers.get(name, "")
    added = ""
    for c in event.text or "":
        if can_add_char(input_type, buf + added, c, name):
            added += c
    if added:
        editor._text_input_buffers[name] = buf + added
    return True



"""Демо hybrid-режима: Kivy UI + встроенная область игры SpritePro.

Запуск:
    python spritePro/demoGames/kivy_hybrid_demo.py

Нужно:
    pip install kivy
"""

from __future__ import annotations

import sys
from pathlib import Path

current_dir = Path(__file__).parent
parent_dir = current_dir.parent.parent
sys.path.insert(0, str(parent_dir))

from kivy.app import App  # noqa: E402
from kivy.uix.boxlayout import BoxLayout  # noqa: E402
from kivy.uix.button import Button  # noqa: E402
from kivy.uix.label import Label  # noqa: E402

import spritePro as s  # noqa: E402


class HybridDemoScene(s.Scene):
    def __init__(self) -> None:
        super().__init__()
        self.player = s.Sprite("", (90, 90), s.WH_C, speed=6, scene=self)
        self.player.set_rect_shape((90, 90), (90, 210, 255), border_radius=24)
        self.info = s.TextSprite(
            "Touch or hold mouse inside game area",
            20,
            (230, 230, 230),
            (20, 20),
            anchor=s.Anchor.TOP_LEFT,
            scene=self,
        )

    def update(self, dt: float) -> None:
        if s.input.is_mouse_pressed(1):
            self.player.set_position(s.input.mouse_pos)


def build_root(game_widget):
    root = BoxLayout(orientation="vertical", spacing=8, padding=8)

    header = BoxLayout(size_hint_y=None, height=56, spacing=8)
    back_button = Button(text="Back", size_hint_x=None, width=120)
    title = Label(text="Hybrid Kivy UI + SpritePro Game")

    def stop_app(*_args) -> None:
        app = App.get_running_app()
        if app is not None:
            app.stop()

    back_button.bind(on_release=stop_app)
    header.add_widget(back_button)
    header.add_widget(title)

    root.add_widget(header)
    root.add_widget(game_widget)
    return root


def main() -> None:
    s.run_kivy_hybrid(
        scene=HybridDemoScene,
        root_builder=build_root,
        size=(1000, 700),
        title="SpritePro Hybrid Demo",
        fill_color=(20, 20, 30),
    )


if __name__ == "__main__":
    main()



"""Нижняя панель статуса и элементов управления."""

from typing import Tuple

import pygame

from . import layouts
from . import theme


def _fit_status_line(font: pygame.font.Font, text: str, max_width: int) -> str:
    if max_width <= 0 or font.size(text)[0] <= max_width:
        return text
    ellipsis = "..."
    left, right = 0, len(text)
    best = ellipsis
    while left <= right:
        mid = (left + right) // 2
        candidate = text[:mid].rstrip() + ellipsis
        if font.size(candidate)[0] <= max_width:
            best = candidate
            left = mid + 1
        else:
            right = mid - 1
    return best


def _render_status_toast(editor) -> None:
    if editor.status_message_timer <= 0:
        return
    viewport = editor._get_viewport_rect()
    max_width = max(180, min(viewport.width - 24, 520))
    text_value = _fit_status_line(editor.font, editor.status_message, max_width - 20)
    is_error = any(
        token in editor.status_message.lower() for token in ("failed", "error", "invalid")
    )
    bg_color = (56, 40, 40) if is_error else (34, 36, 44)
    border_color = (190, 95, 95) if is_error else theme.COLORS["ui_input_border"]
    text_color = (255, 210, 210) if is_error else (230, 230, 238)
    text_surface = editor.font.render(text_value, True, text_color)
    toast_rect = pygame.Rect(0, 0, text_surface.get_width() + 20, text_surface.get_height() + 10)
    toast_rect.midtop = (viewport.centerx, viewport.y + 10)
    pygame.draw.rect(editor.screen, bg_color, toast_rect, border_radius=6)
    pygame.draw.rect(editor.screen, border_color, toast_rect, 1, border_radius=6)
    editor.screen.blit(
        text_surface,
        (toast_rect.x + 10, toast_rect.y + (toast_rect.height - text_surface.get_height()) // 2),
    )


def _draw_slider(
    screen: pygame.Surface,
    rect: pygame.Rect,
    value: float,
    min_value: float,
    max_value: float,
    accent: Tuple[int, int, int],
) -> None:
    pygame.draw.rect(screen, theme.COLORS["ui_slider_track"], rect, border_radius=4)
    ratio = 0.0 if max_value <= min_value else (value - min_value) / (max_value - min_value)
    ratio = max(0.0, min(1.0, ratio))
    fill_width = int(rect.width * ratio)
    if fill_width > 0:
        pygame.draw.rect(
            screen,
            accent,
            pygame.Rect(rect.x, rect.y, fill_width, rect.height),
            border_radius=4,
        )
    thumb_x = rect.x + int(rect.width * ratio)
    thumb_rect = pygame.Rect(thumb_x - 3, rect.y - 2, 6, rect.height + 4)
    pygame.draw.rect(screen, theme.COLORS["ui_slider_thumb"], thumb_rect, border_radius=3)


def render(editor) -> None:
    screen = editor.screen
    font = editor.font
    colors = editor.colors
    w = editor.width
    h = editor.height
    bottom_h = theme.UI_BOTTOM_HEIGHT
    bar_top = h - bottom_h
    bar_rect = pygame.Rect(0, bar_top, w, bottom_h)
    pygame.draw.rect(screen, colors["ui_bg"], bar_rect)
    pygame.draw.line(
        screen,
        colors["ui_border"],
        (0, bar_top),
        (w, bar_top),
        1,
    )
    pad = theme.STATUSBAR_TOP_PADDING

    mouse_text = font.render(
        f"X: {editor.mouse_world_pos.x:.0f}  Y: {editor.mouse_world_pos.y:.0f}",
        True,
        colors["ui_text"],
    )
    screen.blit(mouse_text, (10, bar_top + pad))

    slider_y = bar_top + pad + 3
    slider_h = theme.STATUSBAR_SLIDER_HEIGHT
    sw = theme.STATUSBAR_SLIDER_WIDTH
    gap = theme.STATUSBAR_SLIDER_GAP
    zoom_slider = pygame.Rect(max(180, w // 2 - 180), slider_y, sw, slider_h)
    grid_slider = pygame.Rect(zoom_slider.right + gap, slider_y, sw, slider_h)
    editor._statusbar_controls = {}
    editor._statusbar_controls["zoom"] = zoom_slider
    editor._statusbar_controls["grid"] = grid_slider

    zoom_label = font.render("Zoom", True, colors["ui_text"])
    grid_label = font.render("Grid", True, colors["ui_text"])
    screen.blit(zoom_label, (zoom_slider.x - 40, bar_top + pad - 1))
    screen.blit(grid_label, (grid_slider.x - 34, bar_top + pad - 1))

    _draw_slider(
        screen,
        zoom_slider,
        editor.zoom,
        editor.min_zoom,
        editor.max_zoom,
        colors["ui_accent"],
    )
    _draw_slider(
        screen,
        grid_slider,
        float(editor.scene.grid_size),
        float(editor.min_grid_size),
        float(editor.max_grid_size),
        (150, 130, 255),
    )

    snap_rect = pygame.Rect(
        grid_slider.right + 14,
        bar_top + pad,
        theme.STATUSBAR_SNAP_WIDTH,
        theme.STATUSBAR_SNAP_HEIGHT,
    )
    labels_rect = pygame.Rect(
        snap_rect.right + 6,
        bar_top + pad,
        theme.STATUSBAR_LABELS_WIDTH,
        theme.STATUSBAR_LABELS_HEIGHT,
    )
    editor._statusbar_controls["snap"] = snap_rect
    editor._statusbar_controls["labels"] = labels_rect
    snap_color = colors["ui_accent"] if editor.scene.snap_to_grid else (55, 55, 62)
    snap_fg = theme.COLORS["ui_selected_bg"] if editor.scene.snap_to_grid else colors["ui_text"]
    pygame.draw.rect(screen, snap_color, snap_rect, border_radius=3)
    snap_text = font.render(
        "Snap ON" if editor.scene.snap_to_grid else "Snap OFF",
        True,
        snap_fg,
    )
    screen.blit(snap_text, (snap_rect.centerx - snap_text.get_width() // 2, snap_rect.y + 2))

    labels_color = colors["ui_accent"] if editor.scene.grid_labels_visible else (55, 55, 62)
    labels_fg = (
        theme.COLORS["ui_selected_bg"] if editor.scene.grid_labels_visible else colors["ui_text"]
    )
    pygame.draw.rect(screen, labels_color, labels_rect, border_radius=3)
    labels_text = font.render(
        "Labels ON" if editor.scene.grid_labels_visible else "Labels OFF",
        True,
        labels_fg,
    )
    screen.blit(
        labels_text, (labels_rect.centerx - labels_text.get_width() // 2, labels_rect.y + 2)
    )

    status_strip = layouts.pad(
        pygame.Rect(0, bar_top, w, bottom_h),
        left=0,
        top=pad,
        right=10,
        bottom=4,
    )
    status_strip.height = 18
    status_right_rects = layouts.row_rects(
        status_strip,
        [theme.STATUSBAR_INPUT_WIDTH, theme.STATUSBAR_INPUT_WIDTH],
        gap=theme.STATUSBAR_INPUT_GAP,
        align="right",
    )
    zoom_input_rect = status_right_rects[0]
    grid_input_rect = status_right_rects[1]
    editor._statusbar_controls["zoom_input"] = zoom_input_rect
    editor._statusbar_controls["grid_input"] = grid_input_rect

    _draw_status_input_box(
        editor,
        "zoom_input",
        zoom_input_rect,
        f"{editor.zoom * 100:.0f}",
        "%",
    )
    _draw_status_input_box(
        editor,
        "grid_input",
        grid_input_rect,
        str(editor.scene.grid_size),
        "px",
    )

    _render_status_toast(editor)


def _draw_status_input_box(
    editor,
    name: str,
    rect: pygame.Rect,
    value_display: str,
    suffix: str,
) -> None:
    active = editor._active_text_input == name
    bg = theme.COLORS["ui_input_bg_active"] if active else (45, 45, 52)
    pygame.draw.rect(editor.screen, bg, rect, border_radius=3)
    pygame.draw.rect(editor.screen, editor.colors["ui_border"], rect, 1, border_radius=3)
    if active:
        text = editor._text_input_buffers.get(name, "")
        display = f"{text}|"
    else:
        display = f"{value_display}{suffix}"
    text_surface = editor.font.render(display, True, editor.colors["ui_text"])
    editor.screen.blit(text_surface, (rect.x + 4, rect.y + 2))


def handle_click(editor, pos) -> bool:
    if pos.y < editor.height - theme.UI_BOTTOM_HEIGHT:
        return False
    controls = editor._statusbar_controls
    zoom_rect = controls.get("zoom")
    grid_rect = controls.get("grid")
    snap_rect = controls.get("snap")
    labels_rect = controls.get("labels")
    zoom_input_rect = controls.get("zoom_input")
    grid_input_rect = controls.get("grid_input")

    if zoom_input_rect and zoom_input_rect.collidepoint(pos.x, pos.y):
        editor._activate_text_input("zoom_input", f"{editor.zoom * 100:.0f}", "float")
        return True
    if grid_input_rect and grid_input_rect.collidepoint(pos.x, pos.y):
        editor._activate_text_input(
            "grid_input",
            str(editor.scene.grid_size),
            "int",
            float(editor.min_grid_size),
            float(editor.max_grid_size),
        )
        return True
    if editor._active_text_input:
        editor._deactivate_text_input(apply=True)
    if zoom_rect and zoom_rect.collidepoint(pos.x, pos.y):
        editor._active_slider = "zoom"
        editor._update_active_slider(pos.x)
        return True
    if grid_rect and grid_rect.collidepoint(pos.x, pos.y):
        editor._active_slider = "grid"
        editor._update_active_slider(pos.x)
        return True
    if snap_rect and snap_rect.collidepoint(pos.x, pos.y):
        editor._toggle_scene_setting("snap_to_grid")
        editor._save_state()
        return True
    if labels_rect and labels_rect.collidepoint(pos.x, pos.y):
        editor._toggle_scene_setting("grid_labels_visible")
        editor._save_state()
        return True
    return False



"""Asset hot-reloading utilities for SpritePro."""

from __future__ import annotations

from typing import Callable, Optional, Dict, List, Any
from pathlib import Path

import spritePro


class AssetWatcher:
    """Следит за изменениями файлов и перезагружает ресурсы.

    Использует Polling для отслеживания изменений (не требует watchdog).

    Example:
        >>> watcher = AssetWatcher()
        >>> watcher.watch("assets/images", on_reload=lambda: print("Textures reloaded!"))
        >>> spritePro.register_update_object(watcher)
    """

    def __init__(self, poll_interval: float = 1.0):
        """Инициализирует watcher.

        Args:
            poll_interval (float, optional): Интервал опроса в секундах. По умолчанию 1.0.
        """
        self.poll_interval = poll_interval
        self._watched_paths: Dict[str, Dict[str, Any]] = {}
        self._last_modified: Dict[str, float] = {}
        self._running = False
        self._callbacks: Dict[str, List[Callable]] = {}

    def watch(
        self,
        path: str,
        extensions: Optional[List[str]] = None,
        on_reload: Optional[Callable] = None,
        recursive: bool = True,
    ) -> str:
        """Добавляет путь для отслеживания.

        Args:
            path (str): Путь к папке или файлу.
            extensions (List[str], optional): Список расширений для отслеживания (напр. ['.png', '.jpg']).
                Если None - отслеживаются все файлы. По умолчанию None.
            on_reload (Callable, optional): Колбэк при перезагрузке. По умолчанию None.
            recursive (bool, optional): Рекурсивно отслеживать вложенные папки. По умолчанию True.

        Returns:
            str: ID наблюдателя.
        """
        if extensions is None:
            extensions = []

        path = str(Path(path).resolve())
        watcher_id = f"{path}:{','.join(extensions)}"

        self._watched_paths[watcher_id] = {
            "path": path,
            "extensions": [e.lower() for e in extensions],
            "recursive": recursive,
        }

        if on_reload:
            if watcher_id not in self._callbacks:
                self._callbacks[watcher_id] = []
            self._callbacks[watcher_id].append(on_reload)

        self._scan_files(watcher_id)
        return watcher_id

    def unwatch(self, watcher_id: str) -> None:
        """Удаляет наблюдатель.

        Args:
            watcher_id (str): ID наблюдателя.
        """
        self._watched_paths.pop(watcher_id, None)
        self._callbacks.pop(watcher_id, None)
        for key in list(self._last_modified.keys()):
            if key.startswith(watcher_id):
                del self._last_modified[key]

    def _scan_files(self, watcher_id: str) -> None:
        """Сканирует файлы и обновляет времена модификации."""
        info = self._watched_paths.get(watcher_id)
        if not info:
            return

        path = Path(info["path"])
        extensions = info["extensions"]
        recursive = info["recursive"]

        if not path.exists():
            return

        pattern = "**/*" if recursive else "*"

        for file_path in path.glob(pattern):
            if file_path.is_file():
                if extensions and file_path.suffix.lower() not in extensions:
                    continue
                key = f"{watcher_id}:{file_path}"
                try:
                    self._last_modified[key] = file_path.stat().st_mtime
                except OSError:
                    pass

    def update(self, dt: Optional[float] = None) -> None:
        """Проверяет изменения файлов.

        Args:
            dt (Optional[float], optional): Delta time (не используется).
        """
        for watcher_id in list(self._watched_paths.keys()):
            self._check_changes(watcher_id)

    def _check_changes(self, watcher_id: str) -> None:
        """Проверяет изменения для одного наблюдателя."""
        info = self._watched_paths.get(watcher_id)
        if not info:
            return

        path = Path(info["path"])
        extensions = info["extensions"]
        recursive = info["recursive"]

        if not path.exists():
            return

        pattern = "**/*" if recursive else "*"
        changed = False

        current_files: Dict[str, float] = {}

        for file_path in path.glob(pattern):
            if file_path.is_file():
                if extensions and file_path.suffix.lower() not in extensions:
                    continue

                key = f"{watcher_id}:{file_path}"
                try:
                    mtime = file_path.stat().st_mtime
                    current_files[key] = mtime

                    if key not in self._last_modified:
                        changed = True
                    elif self._last_modified[key] < mtime:
                        changed = True
                        self._last_modified[key] = mtime
                except OSError:
                    pass

        removed_keys = set(self._last_modified.keys()) - set(current_files.keys())
        for key in removed_keys:
            if key.startswith(watcher_id):
                changed = True
                del self._last_modified[key]

        if changed:
            self._reload(watcher_id)

    def _reload(self, watcher_id: str) -> None:
        """Вызывает колбэки при изменении."""
        try:
            import spritePro as sp

            sp.debug_log_info(f"[AssetWatcher] Reloading: {watcher_id}")
        except Exception:
            pass

        if watcher_id in self._callbacks:
            for callback in self._callbacks[watcher_id]:
                try:
                    callback()
                except Exception as e:
                    try:
                        import spritePro as sp

                        sp.debug_log_error(f"[AssetWatcher] Callback error: {e}")
                    except Exception:
                        pass

    def reload_textures(self) -> None:
        """Перезагружает все текстуры в кэше."""
        from .resources import resource_cache

        try:
            resource_cache.clear_textures()
            import spritePro as sp

            sp.debug_log_info("[AssetWatcher] Textures reloaded")
        except Exception as e:
            try:
                import spritePro as sp

                sp.debug_log_error(f"[AssetWatcher] Texture reload failed: {e}")
            except Exception:
                pass

    def reload_sounds(self) -> None:
        """Перезагружает все звуки."""
        from .resources import resource_cache

        try:
            resource_cache.clear_sounds()
            import spritePro as sp

            sp.debug_log_info("[AssetWatcher] Sounds reloaded")
        except Exception as e:
            try:
                import spritePro as sp

                sp.debug_log_error(f"[AssetWatcher] Sound reload failed: {e}")
            except Exception:
                pass

    def reload_all(self) -> None:
        """Перезагружает все ресурсы."""
        self.reload_textures()
        self.reload_sounds()


class HotReloadManager:
    """Менеджер для автоматического hot-reload с удобным API."""

    def __init__(self):
        self.watcher = AssetWatcher()
        self._auto_registered = False

    def watch_textures(
        self,
        folder: str = "assets/images",
        extensions: Optional[List[str]] = None,
    ) -> "HotReloadManager":
        """Наблюдает за текстурами.

        Args:
            folder (str): Папка с текстурами.
            extensions (List[str], optional): Расширения.

        Returns:
            HotReloadManager: self для цепочки.
        """
        if extensions is None:
            extensions = [".png", ".jpg", ".jpeg", ".bmp", ".gif"]
        self.watcher.watch(folder, extensions=extensions, on_reload=self.watcher.reload_textures)
        return self

    def watch_sounds(
        self,
        folder: str = "assets/sounds",
        extensions: Optional[List[str]] = None,
    ) -> "HotReloadManager":
        """Наблюдает за звуками.

        Args:
            folder (str): Папка со звуками.
            extensions (List[str], optional): Расширения.

        Returns:
            HotReloadManager: self для цепочки.
        """
        if extensions is None:
            extensions = [".wav", ".ogg", ".mp3"]
        self.watcher.watch(folder, extensions=extensions, on_reload=self.watcher.reload_sounds)
        return self

    def watch_custom(
        self,
        folder: str,
        extensions: Optional[List[str]] = None,
        callback: Optional[Callable] = None,
    ) -> "HotReloadManager":
        """Наблюдает за пользовательской папкой.

        Args:
            folder (str): Папка для отслеживания.
            extensions (List[str], optional): Расширения.
            callback (Callable, optional): Колбэк при изменении.

        Returns:
            HotReloadManager: self для цепочки.
        """
        self.watcher.watch(folder, extensions=extensions, on_reload=callback)
        return self

    def start(self) -> None:
        """Запускает автоматическое обновление."""
        if not self._auto_registered:
            try:
                spritePro.register_update_object(self.watcher)
                self._auto_registered = True
            except Exception:
                pass

    def stop(self) -> None:
        """Останавливает обновление."""
        if self._auto_registered:
            try:
                spritePro.unregister_update_object(self.watcher)
            except Exception:
                pass


_hot_reload_manager: Optional[HotReloadManager] = None


def get_hot_reload_manager() -> HotReloadManager:
    """Получает глобальный менеджер hot-reload."""
    global _hot_reload_manager
    if _hot_reload_manager is None:
        _hot_reload_manager = HotReloadManager()
    return _hot_reload_manager



from __future__ import annotations

import os
import time
from typing import TYPE_CHECKING

import pygame
from pygame.math import Vector2

from .ui import hierarchy as ui_hierarchy
from .ui import input_handling as ui_input
from .ui import inspector as ui_inspector
from .ui import statusbar as ui_statusbar
from .ui import theme as ui_theme
from .ui import toolbar as ui_toolbar

if TYPE_CHECKING:
    from .editor import SpriteEditor


def _select_tool(editor: "SpriteEditor", index: int) -> None:
    editor.current_tool = editor._toolbar_tools_list[index][0]


def _is_select_tool(editor: "SpriteEditor") -> bool:
    return editor.current_tool == editor._toolbar_tools_list[0][0]


def handle_events(editor: "SpriteEditor") -> None:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            editor.running = False
        elif event.type == pygame.VIDEORESIZE:
            editor.width, editor.height = event.w, event.h
            editor.screen = pygame.display.set_mode((editor.width, editor.height), pygame.RESIZABLE)
            editor.settings_window.rect.center = (editor.width // 2, editor.height // 2)
        elif event.type == pygame.KEYDOWN:
            handle_keydown(editor, event)
        elif event.type == pygame.TEXTINPUT:
            ui_input.handle_text_input_text(editor, event)
        elif event.type == pygame.MOUSEBUTTONDOWN:
            handle_mousedown(editor, event)
        elif event.type == pygame.MOUSEBUTTONUP:
            handle_mouseup(editor, event)
        elif event.type == pygame.MOUSEWHEEL:
            handle_mousewheel(editor, event)
        elif event.type == pygame.DROPFILE:
            handle_dropfile(editor, event)


def handle_keydown(editor: "SpriteEditor", event: pygame.event.Event) -> None:
    if ui_input.handle_text_input_keydown(editor, event):
        return

    keys = pygame.key.get_pressed()
    ctrl_pressed = keys[pygame.K_LCTRL] or keys[pygame.K_RCTRL]
    shift_pressed = keys[pygame.K_LSHIFT] or keys[pygame.K_RSHIFT]

    if ctrl_pressed:
        if event.key == pygame.K_n:
            editor._new_scene()
            return
        if event.key == pygame.K_o:
            editor._load_scene()
            return
        if event.key == pygame.K_z:
            editor.undo()
            return
        if event.key == pygame.K_y:
            editor.redo()
            return
        if event.key == pygame.K_c:
            editor.copy_selected()
            return
        if event.key == pygame.K_v:
            editor.copy_selected()
            return
        if event.key == pygame.K_t and shift_pressed:
            editor.add_text()
            editor._set_status("Text object created")
            return
        if event.key == pygame.K_s:
            if shift_pressed:
                editor._save_scene_as()
            else:
                editor._save_scene()
            return

    if event.key == pygame.K_v:
        _select_tool(editor, 0)
    elif event.key == pygame.K_g:
        _select_tool(editor, 1)
    elif event.key == pygame.K_r:
        _select_tool(editor, 2)
    elif event.key == pygame.K_t:
        _select_tool(editor, 3)
    elif event.key == pygame.K_F1:
        editor.window_manager.toggle("settings")
    elif event.key == pygame.K_F5:
        editor._run_project()
    elif event.key == pygame.K_DELETE or event.key == pygame.K_BACKSPACE:
        editor.delete_selected()
    elif event.key == pygame.K_f:
        editor._frame_selection()
    elif event.key == pygame.K_ESCAPE:
        ui_toolbar.close_menu(editor)
        ui_hierarchy.close_context_menu(editor)
        editor.deselect_all()


def handle_mousedown(editor: "SpriteEditor", event: pygame.event.Event) -> None:
    editor.mouse_pos = Vector2(event.pos)
    editor.mouse_world_pos = editor.screen_to_world(editor.mouse_pos)

    if event.button == 1:
        editor.mouse_pressed = True
        editor._transform_changed_during_drag = False

        if editor._active_text_input and not editor._click_in_any_text_input(event.pos):
            editor._deactivate_text_input(apply=True)

        if ui_toolbar.handle_click(editor, editor.mouse_pos):
            return
        if ui_hierarchy.handle_menu_click(editor, editor.mouse_pos):
            return
        if editor.window_manager.handle_click((int(editor.mouse_pos.x), int(editor.mouse_pos.y))):
            return
        if ui_statusbar.handle_click(editor, editor.mouse_pos):
            return

        if editor.mouse_pos.x >= editor.width - editor.ui_right_width and ui_inspector.handle_click(
            editor, editor.mouse_pos
        ):
            return

        if editor.mouse_pos.x <= editor.ui_left_width:
            obj = ui_hierarchy.handle_click(editor, editor.mouse_pos)
            if obj is not None:
                ui_hierarchy.close_context_menu(editor)
                now = time.time()
                is_double = obj == editor._last_hierarchy_click_obj and (
                    now - editor._last_hierarchy_click_time
                ) < ui_theme.DOUBLE_CLICK_INTERVAL
                editor._last_hierarchy_click_time = now
                editor._last_hierarchy_click_obj = obj
                if is_double and obj != "__camera__":
                    editor._frame_selection()
                    return
                if obj == "__camera__":
                    editor.camera_selected = True
                    editor.selected_objects.clear()
                else:
                    keys = pygame.key.get_pressed()
                    editor.select_object(
                        obj, add_to_selection=keys[pygame.K_LSHIFT] or keys[pygame.K_RSHIFT]
                    )
                return

        viewport = editor._get_viewport_rect()
        if viewport.collidepoint(editor.mouse_pos):
            if editor._camera_preview_copy_rect and editor._camera_preview_copy_rect.collidepoint(
                editor.mouse_pos.x, editor.mouse_pos.y
            ):
                editor._copy_scene_camera_to_game()
                return
            obj = editor.get_object_at(editor.mouse_world_pos)
            if obj:
                ui_hierarchy.close_context_menu(editor)
                keys = pygame.key.get_pressed()
                editor.select_object(
                    obj, add_to_selection=keys[pygame.K_LSHIFT] or keys[pygame.K_RSHIFT]
                )
            elif _is_select_tool(editor):
                editor.deselect_all()

            editor.drag_start = Vector2(editor.mouse_world_pos.x, editor.mouse_world_pos.y)
            editor.camera_drag_start = Vector2(event.pos)
            editor._capture_drag_state()
            editor.mouse_dragging = True

    elif event.button == 2:
        ui_toolbar.close_menu(editor)
        ui_hierarchy.close_context_menu(editor)
        editor.mouse_pressed = True
        editor.camera_drag_start = Vector2(event.pos)

    elif event.button == 3:
        ui_toolbar.close_menu(editor)
        if editor.mouse_pos.x <= editor.ui_left_width:
            obj = ui_hierarchy.handle_click(editor, editor.mouse_pos)
            if obj not in (None, "__camera__"):
                editor.select_object(obj)
                ui_hierarchy.open_context_menu(editor, obj, editor.mouse_pos)
                return
            ui_hierarchy.close_context_menu(editor)
            return
        ui_hierarchy.close_context_menu(editor)
        editor.mouse_pressed = True
        editor.camera_drag_start = Vector2(event.pos)


def handle_mouseup(editor: "SpriteEditor", event: pygame.event.Event) -> None:
    if event.button == 1:
        if editor._active_slider is not None:
            editor._active_slider = None
            editor._save_state()
        elif editor.mouse_dragging and editor._transform_changed_during_drag:
            editor._save_state()

    editor.mouse_pressed = False
    editor.mouse_dragging = False
    editor._drag_object_starts.clear()
    editor._transform_changed_during_drag = False


def handle_mousewheel(editor: "SpriteEditor", event: pygame.event.Event) -> None:
    mouse_pos = Vector2(pygame.mouse.get_pos())
    if ui_hierarchy.handle_wheel(editor, event.y):
        return
    viewport_rect = editor._get_viewport_rect()
    if not viewport_rect.collidepoint(mouse_pos.x, mouse_pos.y):
        return
    zoom_factor = 1.1 if event.y > 0 else 1 / 1.1
    editor._set_zoom(editor.zoom * zoom_factor, Vector2(pygame.mouse.get_pos()))


def handle_dropfile(editor: "SpriteEditor", event: pygame.event.Event) -> None:
    filepath = event.file
    ext = os.path.splitext(filepath)[1].lower()
    if ext in [".png", ".jpg", ".jpeg", ".bmp", ".gif"]:
        world_pos = editor.screen_to_world(Vector2(pygame.mouse.get_pos()))
        editor.add_sprite(filepath, world_pos)
